
/*
 *  SIMD Smith-Waterman function for x86 platform
 *
 *  Technology: Intel MMX/SSE/SSE2
 *  Algorithm:  Horizontal
 *  Precision:  Byte, Word
 *
 *  Copyright (C) 2003 Arpith Chacko Jacob
 *  Email: jarpith@acm.org
 *  Beta release 1.0.0.2
 *
 *  Start date:  28 june 2003
 *  Last Update: 08 september 2003
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#ifdef DEBUG
	#include <time.h>
#endif

#include "sw_simd_x86.h"
#include "sw_horiz.h"

/*
 *  Function to search a sequence database using the Smith-Waterman algorithm
 *  implemented using SIMD technology on Intel x86.
 *
 *  Inputs: 
 *  Returns: 
 *
 */
int sw_horiz (int simd_tech, int precision, int argc, char *argv[])
{
    FILE *file_d, *file_q, *file_s;
    unsigned int i, j, k;

	unsigned char *p1, *p2;
	unsigned short *p3;

    char *sim_mat;
    unsigned char alph_size, blank_char;
    char bias;
    char query_id[ID_LEN], seq_id[ID_LEN];
    unsigned char *query, *seq;
	unsigned char *query_profile_c[2];
	unsigned short *query_profile_s[2];
    unsigned short *fih[2];
    unsigned char gap_init;
    unsigned char gap_ext;
    unsigned int m, n;

    int score;

	int search_type;
	unsigned int simd_byte_vlen, simd_word_vlen;

	/* Function pointers to assembly routines according to SIMD tech selected */
	int (*simd_horiz_b_ptr) (char *, int, char *, int, char *, int, int, int, char *);

	int (*simd_horiz_w_ptr) (char *, int, char *, int, short *, int, int, int, short *);

#ifdef	DEBUG
    double ovr_time = 0, sw_time = 0;
    unsigned int total_seq = 0;
    unsigned int total_n = 0;
    float _khz;				// CPU speed
    clock_t pclock;

	unsigned int slo, shi, elo, ehi;
#endif

	printf ("Horizontal approach...\n");

	switch (simd_tech) {
		case SW_SIMD_MMX:
			simd_horiz_b_ptr = &sw_simd_mmx_horiz_b;
			simd_horiz_w_ptr = &sw_simd_mmx_horiz_w;
			simd_byte_vlen = SW_SIMD_MMX_BYTE_VLEN;
			simd_word_vlen = SW_SIMD_MMX_WORD_VLEN;
			printf ("MMX Technology...\n");
			break;
		case SW_SIMD_SSE:
			simd_horiz_b_ptr = &sw_simd_sse_horiz_b;
			simd_horiz_w_ptr = &sw_simd_sse_horiz_w;
			simd_byte_vlen = SW_SIMD_SSE_BYTE_VLEN;
			simd_word_vlen = SW_SIMD_SSE_WORD_VLEN;
			printf ("SSE Technology...\n");
			break;
		case SW_SIMD_SSE2:
			simd_horiz_b_ptr = &sw_simd_sse2_horiz_b;
			simd_horiz_w_ptr = &sw_simd_sse2_horiz_w;
			simd_byte_vlen = SW_SIMD_SSE2_BYTE_VLEN;
			simd_word_vlen = SW_SIMD_SSE2_WORD_VLEN;
			printf ("SSE2 Technology...\n");
			break;
		default:
			return -1;
	}

	switch (precision) {
		case SW_SIMD_BYTE:
			printf ("Byte precision...\n");
			break;
		case SW_SIMD_WORD:
			printf ("Word precision...\n");
			break;
		case SW_SIMD_BYTE | SW_SIMD_WORD:
			printf ("Byte/Word precision...\n");
			break;
		default:
			return -1;
	}

	// Set Protein/DNA search
	if ('A' == argv[4][0] || 'a' == argv[4][0]) {
		printf ("Performing protein search...\n");
		search_type = SW_PROTEIN_SEARCH;
		alph_size = 24;		// One additional dummy character
		blank_char = 23;
	} else {
		printf ("Performing DNA search...\n");
		search_type = SW_DNA_SEARCH;
		alph_size = 5;		// One additional dummy character
		blank_char = 4;
	}

    // Open files
    file_d = fopen (argv[5], "r");
    file_q = fopen (argv[6], "r");
    file_s = fopen (argv[7], "r");

    if (!file_d || !file_q || !file_s) {
        fprintf (stderr, "File does not exist\n");
        return -1;
    }

#ifdef	DEBUG
    // measure KHZ
    pclock = clock ();

    __asm__ __volatile__ ( "cpuid\n\t"
	  "rdtsc\n\t"
	  "movl %%eax, %0"
	  : "=g" (slo)
	  : 
	  : "%eax", "%ebx", "%ecx", "%edx", "memory" );

    while (clock () - pclock < CLOCKS_PER_SEC) _khz++; 

    __asm__ __volatile__ ( "cpuid\n\t"
	  "rdtsc\n\t"
	  "sub %1, %%eax\n\t"
	  "mov %%eax, %0"
	  : "=g" (slo)
	  : "g" (slo)
	  : "%eax", "%ebx", "%ecx", "%edx", "memory" );


     _khz = (float) slo / 1000;
    printf ("MHZ on this machine: %f\n", _khz / 1000);

    // Whole program timing measurement
    pclock = clock ();
#endif

    // Gap parameters
    gap_init = atoi (argv[8]);
    gap_ext = atoi (argv[9]);

    // Allocate memory for sim_mat
    sim_mat = (char *) malloc (alph_size * alph_size * sizeof (char));
    if (!sim_mat) {
        fprintf (stderr, "Cannot allocate memory for sim_mat\n");
        return -1;
    }

    // Read similarity matrix and determine bias
    bias = 0x7f;
    for (i = 0; i < alph_size - 1; i++) {
        for (j = 0; j < alph_size - 1; j++) {
            if (fscanf (file_s, "%d ", (int *) (sim_mat + i * alph_size + j)) == EOF) {
                fprintf (stderr, "Error in similarity matrix\n");
		        free (sim_mat);
                return -1;
            }

            if (*(sim_mat + i * alph_size + j) < bias) {
            	bias = *(sim_mat + i * alph_size + j);
            }
        }
    }
    fclose (file_s);

	// BIAS similarity matrix so that no value is < 0
	bias = -bias;
    for (i = 0; i < alph_size - 1; i++) {
        for (j = 0; j < alph_size - 1; j++) {
            *(sim_mat + i * alph_size + j) += bias;
        }
        // Add dummy character to each row
        *(sim_mat + i * alph_size + alph_size - 1) = 0;
    }

    // Add dummy character row at the end of the matrix
    for (j = 0; j < alph_size; j++) {
        *(sim_mat + (alph_size - 1) * alph_size + j) = 0;
    }

#ifdef	DEBUG
	printf ("Bias = %d\n", bias);
#endif

/*    for (i = 0; i < alph_size; i++) {
        for (j = 0; j < alph_size; j++) {
            printf ("%d ", *(sim_mat + i * alph_size + j));
        }
        printf ("\n");
    }
*/

    // Read query
    query = (char *) malloc (QBUF_LEN * sizeof (char));
    if (!query) {
        fprintf (stderr, "Cannot allocate memory for query sequence\n");
        free (sim_mat);
        return -1;
    }

	m = next_seq(file_q, search_type, query_id, query, QBUF_LEN);
	k = m;

	// Adjust to multiple of SIMD_SIZE
	if ((j = (m % simd_byte_vlen))) j = simd_byte_vlen - j;
	for (i = 0; i < j; i++)
		*(query + m + i) = blank_char;
	m += j;

    fclose (file_q);

    printf ("\nQuery: %s\n", query_id);

    // Allocate memory for database sequence
    seq = (char *) malloc (SBUF_LEN * sizeof (char));

    if (!seq) {
        fprintf (stderr, "Cannot allocate memory for database sequence\n");
        free (sim_mat);
        free (query);
        return -1;
    }

    // Allocate memory for computation
    memalign ((char **) fih, simd_byte_vlen, 2 * m * sizeof (short));
	memalign((char **) query_profile_c, simd_byte_vlen, alph_size * m * sizeof(char));
	memalign((char **) query_profile_s, simd_byte_vlen, alph_size * m * sizeof(short));

    if (!fih[0]) {
        fprintf (stderr, "Cannot allocate memory for F() and H()\n");
        free (sim_mat);
        free (query);
        free (seq);
        return -1;
    }

	if (!query_profile_c[0]) {
		fprintf (stderr, "Cannot allocate memory for query profile (char)\n");
        free (sim_mat);
        free (query);
        free (seq);
    	free (fih[0]);
        return -1;
	}

	if (!query_profile_s[0]) {
		fprintf (stderr, "Cannot allocate memory for query profile (short)\n");
        free (sim_mat);
        free (query);
        free (seq);
    	free (fih[0]);
    	free (query_profile_c[0]);
        return -1;
	}

	// Calculate query profile
	p1 = query_profile_c[1];
	p3 = query_profile_s[1];
	p2 = sim_mat;
	for (i = 0; i < alph_size; i++) {
		for (j = 0; j < m; j++) {
			*p1 = *(p2 + *(query + j));
			*p3 = (short) *(p1);
			p1++;
			p3++;
		}
		p2 += alph_size;
	}

	/* Perform S-W searching */
	if ((SW_SIMD_BYTE | SW_SIMD_WORD) == precision) {

		/* Use both precisions as required */
    	while (!feof (file_d)) {

	#ifdef	DEBUG
        	start_counter (&slo, &shi);
	#endif

        	// Read next database sequence
			n = next_seq(file_d, search_type, seq_id, seq, SBUF_LEN);

	#ifdef	DEBUG
    	    ovr_time += end_counter (&slo, &shi, &elo, &ehi);
	        start_counter (&slo, &shi);
	#endif

			// Use byte precision for maximum speed
       		score = simd_horiz_b_ptr (query, m, seq, n, query_profile_c[1],
       								  bias, gap_init, gap_ext, (char *) fih[1]);

			// If at limit, switch to word precision
			if (score >= SW_SIMD_BYTE_SATURATION - bias) {
				// Max score possible = 2 ^ 16 - 1 - bias
       			score = simd_horiz_w_ptr (query, m, seq, n, query_profile_s[1],
       									  bias, gap_init, gap_ext, fih[1]);
			}

	#ifdef	DEBUG
    	    sw_time += end_counter (&slo, &shi, &elo, &ehi);

	        // For average timing
        	total_n += n;
        	total_seq++;
	#endif
        	printf ("Sequence: %s (%d), Score: %d\n", seq_id, total_seq, score);
        
    	}

	} else if (SW_SIMD_WORD == precision) {

		/* Use word precision only */
    	while (!feof (file_d)) {

	#ifdef	DEBUG
        	start_counter (&slo, &shi);
	#endif

        	// Read next database sequence
			n = next_seq(file_d, search_type, seq_id, seq, SBUF_LEN);

	#ifdef	DEBUG
    	    ovr_time += end_counter (&slo, &shi, &elo, &ehi);
	        start_counter (&slo, &shi);
	#endif

   			score = simd_horiz_w_ptr (query, m, seq, n, query_profile_s[1],
   									  bias, gap_init, gap_ext, fih[1]);

	#ifdef	DEBUG
    	    sw_time += end_counter (&slo, &shi, &elo, &ehi);

	        // For average timing
        	total_n += n;
        	total_seq++;
	#endif
        	printf ("Sequence: %s (%d), Score: %d\n", seq_id, total_seq, score);
        
    	}

	} else if (SW_SIMD_BYTE == precision) {

		/* Use byte precision only */
    	while (!feof (file_d)) {

	#ifdef	DEBUG
        	start_counter (&slo, &shi);
	#endif

        	// Read next database sequence
			n = next_seq(file_d, search_type, seq_id, seq, SBUF_LEN);

	#ifdef	DEBUG
    	    ovr_time += end_counter (&slo, &shi, &elo, &ehi);
	        start_counter (&slo, &shi);
	#endif

       		score = simd_horiz_b_ptr (query, m, seq, n, query_profile_c[1],
       								  bias, gap_init, gap_ext, (char *) fih[1]);

	#ifdef	DEBUG
    	    sw_time += end_counter (&slo, &shi, &elo, &ehi);

	        // For average timing
        	total_n += n;
        	total_seq++;
	#endif
        	printf ("Sequence: %s (%d), Score: %d\n", seq_id, total_seq, score);
        
    	}

    }

#ifndef	DEBUG
	switch (simd_tech) {
		case SW_SIMD_MMX:
		case SW_SIMD_SSE:
			/* End of MMX */
			asm ("emms");
			break;
	}
#endif

#ifdef	DEBUG
    pclock = clock () - pclock;

    printf ("TIMING:\n");
    printf ("Cells (adjusted): %d, Cells/second: %f\n", m * total_n, (double) (m * total_n) / ((double) sw_time / (_khz * 1000)));
    printf ("DB seq: %u, DB bases: %u, Query: %u\n", total_seq, total_n, m);
    printf ("PROGRAM: Total %.2f seconds. clock()\n", pclock / (double)CLOCKS_PER_SEC);
    printf ("TOTAL:   Overhead %f ms, SW %f ms.\n", ovr_time / _khz, sw_time / _khz);
    printf ("PERSEQ:  Overhead %f ms, SW %f ms.\n", ovr_time / _khz / total_seq, sw_time / _khz / total_seq);
#endif
    
    fclose (file_d);

    free (fih[0]);
    free (query_profile_c[0]);
    free (query_profile_s[0]);
    free (sim_mat);
    free (query);
    free (seq);

    return 0;	
}

