
/*
 *  SIMD Smith-Waterman functions for x86 platform
 *
 *  Technology: Intel MMX, SSE, SSE2
 *  Algorithms: Horizontal (optimum for large affine gap penalties)
 *              Diagonal (consistent speed for all gap penalties)
 *  Precision:  Byte, Word
 *
 *  Copyright (C) 2003 Arpith Chacko Jacob
 *  Email: jarpith@acm.org
 *  Beta release 1.0.0.2
 *
 *  Start date:  20 june 2003
 *  Last Update: 08 september 2003
 *
 */

#include <stdlib.h>
#include <stdio.h>

#include "sw_simd_x86.h"
#include "sw_horiz.h"
#include "sw_diag.h"

/*
 * Main function
 *
 */
int main (int argc, char *argv[])
{
/*	unsigned int shi, slo, ehi, elo; */
	int type, tech;

    // Check for arguments
    if (argc != 10) {
        fprintf (stderr, "Invalid args\n%s h|d mmx|sse|sse2|auto b|w|bw a|n FASTA_database FASTA_query sim_matrix gap_init gap_ext\n", argv[0]);
        return -1;
    }

	// Get SIMD technology
	if (('m' == argv[2][0] || 'M' == argv[2][0]) &&
				('m' == argv[2][1] || 'M' == argv[2][1]) &&
				('x' == argv[2][2] || 'X' == argv[2][2])) {
		tech = SW_SIMD_MMX;
	} else if (('s' == argv[2][0] || 'S' == argv[2][0]) &&
				('s' == argv[2][1] || 'S' == argv[2][1]) &&
				('e' == argv[2][2] || 'E' == argv[2][2]) &&
				('2' == argv[2][3])) {
		tech = SW_SIMD_SSE2;
	} else if (('s' == argv[2][0] || 'S' == argv[2][0]) &&
				('s' == argv[2][1] || 'S' == argv[2][1]) &&
				('e' == argv[2][2] || 'E' == argv[2][2])) {
		tech = SW_SIMD_SSE;
	} else {
		tech = get_simd_tech ();
	}

	// Get precision
	switch (argv[3][0]) {
		case 'b':
		case 'B':
			if ('w' == argv[3][1] || 'W' == argv[3][1]) {
				type = SW_SIMD_BYTE | SW_SIMD_WORD;
			} else {
				type = SW_SIMD_BYTE;
			}
			break;
		case 'w':
		case 'W':
			type = SW_SIMD_WORD;
			break;
		default:
			type = SW_SIMD_BYTE | SW_SIMD_WORD;
	}		

	// Get approach
	switch (argv[1][0]) {
		case 'h':
		case 'H':
			sw_horiz (tech,
			  		  type,
			  		  argc,
			  		  argv);
			break;
		default:
			sw_diag (tech,
			  		 type,
			  		 argc,
			  		 argv);
	}

/*	// Example of timing measurement
	start_counter (&shi, &slo);
	sleep(9);
	printf ("Time: %f\n", end_counter (&shi, &slo, &ehi, &elo) / 500e6);	// 500 Mhz machine

*/


    return 0;	
}

/*
 *  Function to return aligned memory
 *
 *  Inputs: Array of pointers - ptr[0] = unaligned, ptr[1] = aligned,
 *          Alignment required, Size of memory requested
 *  Returns: Aligned memory location
 *
 */
void *memalign (char *p[2], int alignment, unsigned int size)
{
    register void *ptr = malloc (size + alignment);
    register void *aligned;

    p[0] = p[1] = NULL;

    if (ptr == 0) {
        return NULL;
    }

    /* Align memory */
    if (((int) ptr & (alignment - 1)) == 0) {
        aligned = ptr;
    } else {
        aligned = (void *) (((int) ptr + alignment) & -alignment);
    }

    p[0] = ptr;
    p[1] = aligned;

    return aligned;
}

/*
 *  Function to return next FASTA sequence from file
 *
 *  Inputs: Opened FASTA file ptr, Sequence ID, Sequence, Size of Sequence buffer
 *  Returns: Size of sequence on success, -1 on failure
 *
 */
inline int next_seq (FILE *f, int seq_type, char *id, char *seq, int mem_size)
{
    int i;
    char c;

    // Read query file
    while ((c = fgetc (f)) != '>' && c != EOF);
    // Remove any extra '>'
    while ((c = fgetc (f)) == '>');

    // Read query identifier
    for (i = 0; i < ID_LEN - 1 && c != EOF && c != '\n'; i++) {
        if (c != '\r') 
	    id[i] = c;
        c = fgetc (f);
    }
    id[i] = '\0';

    // Read till end of identifier
    ungetc (c, f);
    while ((c = fgetc (f)) != '\n');
    ungetc (c, f);

    // FASTA annotation line
    while ((c = fgetc (f)) == ';')
        while ((c = fgetc (f)) != '\n');
    ungetc (c, f);

    // Read query sequence
    for (i = 0; (c = fgetc (f)) != EOF && c != '>' && c != ';'; ) {
        if (c && c != '\n' && c != '\r') 
            seq[i++] = (SW_PROTEIN_SEARCH == seq_type) ? AA2int (c) : DNA2int (c);
		
        if (i >= mem_size) {
            mem_size += BUF_INCR;

            if (!realloc (seq, mem_size * sizeof (char))) {
                fprintf (stderr, "Cannot re-allocate memory for query sequence\n");
                return -1;
            }
        }
    }

    if (c != EOF)
        ungetc (c, f);

    return i;
}

/*
 *  Function to convert ascii nucleotide data to numeric values
 *
 *  Inputs: Ascii sequence character
 *  Returns: Integral equivalent
 *
 */
inline char DNA2int (char a)
{

    switch (a) {
        case 'a':
        case 'A': return 0; 
        case 'c':
        case 'C': return 1; 
        case 'g':
        case 'G': return 2; 
        case 't':
        case 'T': return 3; 
        case 'n':
        case 'N': return 4;		// blank
        default:
            fprintf (stderr, "Invalid sequence character\n");
            exit (1);
    }
}

/*
 *  Function to convert ascii amino acid data to numeric values
 *
 *  Inputs: Ascii sequence character
 *  Returns: Integral equivalent
 *
 */
inline char AA2int (char a)
{

    switch (a) {
        case 'a':
        case 'A': return 0; 
        case 'r':
        case 'R': return 1; 
        case 'n':
        case 'N': return 2; 
        case 'd':
        case 'D': return 3; 
        case 'c':
        case 'C': return 4;
        case 'q':
        case 'Q': return 5;
        case 'e':
        case 'E': return 6;
        case 'g':
        case 'G': return 7;
        case 'h':
        case 'H': return 8;
        case 'i':
        case 'I': return 9;
        case 'l':
        case 'L': return 10;
        case 'k':
        case 'K': return 11;
        case 'm':
        case 'M': return 12;
        case 'f':
        case 'F': return 13;
        case 'p':
        case 'P': return 14;
        case 's':
        case 'S': return 15;
        case 't':
        case 'T': return 16;
        case 'w':
        case 'W': return 17;
        case 'y':
        case 'Y': return 18;
        case 'v':
        case 'V': return 19;
        case 'b':
        case 'B': return 20;
        case 'z':
        case 'Z': return 21;
        case 'x':
        case 'X': return 22;
        default:
            fprintf (stderr, "Invalid sequence character\n");
            exit (1);
    }
}

/*
 *  Function to record clock ticks, to measure short time intervals
 *  (x86 specific).
 *		NOTE: Not thread safe. Context-switching may also affect values.
 *            May not be accurate on Windows 98/ME
 *            Call accompanying function to return time interval
 *
 *  Inputs: unsigned int pointer start high, unsigned int pointer start low
 *  Returns: None
 *
 */
inline void start_counter (unsigned int *shi, unsigned int *slo)
{
	/* CPUID makes sure all previous instructions have been executed */
    __asm__ __volatile__ ( "cpuid\n\t"
	  "rdtsc\n\t"
 	  "mov %%edx, %0\n\t"
  	  "mov %%eax, %1"
	  : "=g" (*shi), "=g" (*slo)
	  : 
	  : "%eax", "%ebx", "%ecx", "%edx", "memory" );
}

/*
 *  Function to return time duration in clock ticks
 *  (x86 specific).
 *		NOTE: Not thread safe. Context-switching may also affect values.
 *            May not be accurate on Windows 98/ME
 *            Must call accompanying function to start recording
 *
 *  Inputs: unsigned int pointer start high, unsigned int pointer start low
 *          unsigned int pointer end high, unsigned int pointer end low
 *  Returns: Time interval in clock ticks.
 *           Time in seconds = CLOCK TICKS / CPU CLOCK FREQUENCY IN HERTZ
 *
 */
inline double end_counter (unsigned int *shi, unsigned int *slo,
			  			   unsigned int *ehi, unsigned int *elo)
{
    unsigned int borrow;

    __asm__ __volatile__ ( "cpuid\n\t"
	  "rdtsc\n\t"
 	  "mov %%edx, %0\n\t"
  	  "mov %%eax, %1"
	  : "=g" (*ehi), "=g" (*elo)
	  : 
	  : "%eax", "%ebx", "%ecx", "%edx", "memory" );

    *slo = *elo - *slo;
    borrow = *slo > *elo;
    *shi = *ehi - *shi - borrow;

    return (double) *shi * (1 << 30) * 4 + *slo;
}

/*
 *  Function to return best SIMD technology available on this machine
 *
 *		NOTE: Tested on Intel P3. No idea how it runs on AMD/Cyrix processors.
 *
 *  Inputs: None
 *  Returns: 0 - No SIMD technology detected
 *           1 - MMX
 *           2 - SSE
 *           3 - SSE2
 *
 */
inline int get_simd_tech ()
{
	int simd_type;

	/* Check for SIMD technology */
	__asm__ __volatile__ ( /* Check for CPUID instruction: Toggle bit 21 */
	"pushf\n\t"
	"pop %%eax\n\t"
	"xor $0x200000, %%eax\n\t"
	"push %%eax\n\t"
	"popf\n\t"
	"pushf\n\t"
	"pop %%ebx\n\t"
	"xor %%ebx, %%eax\n\t"
	"jnz no_cpuid\n\t"

	/* Get max eax */
	"xor %%eax, %%eax\n\t"
	"cpuid\n\t"
	"cmp $1, %%eax\n\t"
	"jl no_cpuid\n\t"
	
	/* Check for SSE2/SSE/MMX */
	"mov $1, %%eax\n\t"
	"cpuid\n\t"
	
	/* SSE2 detection */
	"movl %4, %0\n\t"
	"test $0x4000000, %%edx\n\t"
	"jnz done\n\t"

	/* SSE detection */
	"movl %3, %0\n\t"
	"test $0x2000000, %%edx\n\t"
	"jnz done\n\t"

	/* MMX detection */
	"movl %2, %0\n\t"
	"test $0x800000, %%edx\n\t"
	"jnz done\n\t"

	"no_cpuid: movl %1, %0\n\t"
	"done: "
	
	: "=g" (simd_type)
	: "I" (SW_SIMD_NO_SIMD), "I" (SW_SIMD_MMX), "I" (SW_SIMD_SSE),
	  "I" (SW_SIMD_SSE2)
	: "eax", "ebx", "ecx", "edx" );

	return simd_type;
}
